@@ -25,7 +25,7 @@ use Encode 'encode_utf8';
 
 use experimental 'postderef';
 
-our $VERSION = "v0.0.13";
+our $VERSION = "v0.0.14";
 
 use constant {
     DEFAULT_HEARTBEAT_INTERVAL        => 5,
@@ -125,9 +125,7 @@ sub register ($self, $app, $config) {
         };
 
         # register heartbeat every X seconds, or gracefully stop
-        rx_merge(
-            rx_timer(rand($heartbeat_interval), $heartbeat_interval),
-        )->pipe(
+        rx_timer(rand($heartbeat_interval), $heartbeat_interval)->pipe(
             op_take_until($graceful_stop_o),
         )->subscribe(sub {
             try {
@@ -342,18 +340,25 @@ sub register ($self, $app, $config) {
                 my $join_handler = BoardStreams::Registry->get_join($channel_name) or return;
                 $c->$bs_prefix->channel_exists($channel_name) or return;
                 my $attrs = { since_id => $since_id, is_reconnect => int(defined $since_id) };
+
+                my $o = get_listener_observable($c, $channel_name, $bs_prefix);
+                my $s = $o->subscribe();
+
                 my $join_ret = eval { $join_handler->($channel_name, $c, $attrs) };
                 if ($join_ret->$_can('then')) {
                     $join_ret = eval { await $join_ret };
                 }
-                $join_ret or return;
-                my $limit = $join_ret->{limit} // 0;
+                if (! $join_ret) {
+                    $s->unsubscribe();
+                    return;
+                }
+                my $limit = eval { $join_ret->{limit} } // 0;
 
                 my $was_added = BoardStreams::Registry->add_pair($c, $channel_name);
                 if ($was_added) {
-                    my $o = get_listener_observable($c, $channel_name, $bs_prefix);
-                    my $s = $o->subscribe();
                     $c->stash->{subscriptions}{channels}{$channel_name} = $s;
+                } else {
+                    $s->unsubscribe();
                 }
 
                 # NOTE: in the async case in the future, make sure the subscription is still there
